import argparse
from concurrent import futures
import os
import commentjson as json

import numpy as np

import shutil
import time

from common import *
from scenes import scenes_nerf, scenes_image, scenes_sdf, scenes_volume, setup_colored_sdf

from tqdm import tqdm

import grpc
import render_pb2
import render_pb2_grpc

import pyngp as ngp # noqa

def parse_args():
	parser = argparse.ArgumentParser(description="Run neural graphics primitives testbed with additional configuration & output options")

	parser.add_argument("--load_snapshot", default="", help="Load this snapshot before training. recommended extension: .msgpack")

	parser.add_argument("--exposure", default=0.0, type=float, help="Controls the brightness of the image. Positive numbers increase brightness, negative numbers decrease it.")

	parser.add_argument("--screenshot_transforms", default="", help="Path to a nerf style transforms.json from which to save screenshots.")
	parser.add_argument("--screenshot_dir", default="", help="Which directory to output screenshots to.")

	parser.add_argument("--width", "--screenshot_w", type=int, default=760, help="Resolution width of GUI and screenshots.")
	parser.add_argument("--height", "--screenshot_h", type=int, default=428, help="Resolution height of GUI and screenshots.")
	parser.add_argument("--screenshot_spp", type=int, default=16, help="Number of samples per pixel in screenshots.")

	parser.add_argument("--sharpen", default=0, help="Set amount of sharpening applied to NeRF training images.")

	parser.add_argument("--serve", action="store_true")

	args = parser.parse_args()
	return args

class Renderer(render_pb2_grpc.RenderServicer):

	def __init__(self, testbed, width, height, spp):
		self.testbed = testbed
		self.width = width
		self.height = height
		self.spp = spp

	def Render(self, request, context):
		pose = np.matrix(request.data).reshape((4,4))
		testbed.set_nerf_camera_matrix(pose[:-1,:])
		rendered_image = testbed.render(self.width, self.height, self.spp, True)
		print(f"rendered image with tf {request.data}")
		return render_pb2.Image(data=rendered_image)


if __name__ == "__main__":
	args = parse_args()

	mode = ngp.TestbedMode.Nerf

	testbed = ngp.Testbed(mode)
	testbed.nerf.sharpen = float(args.sharpen)
	testbed.exposure = args.exposure

	if args.load_snapshot:
		print("Loading snapshot ", args.load_snapshot)
		testbed.load_snapshot(args.load_snapshot)

	ref_transforms = {}
	if args.screenshot_transforms: # try to load the given file straight away
		print("Screenshot transforms from ", args.screenshot_transforms)
		with open(args.screenshot_transforms) as f:
			ref_transforms = json.load(f)

	testbed.shall_train = False

	testbed.nerf.render_with_camera_distortion = True

	if ref_transforms:
		testbed.fov_axis = 0
		testbed.fov = ref_transforms["camera_angle_x"] * 180 / np.pi

		screenshot_frames = range(len(ref_transforms["frames"]))
		print(screenshot_frames)
		for idx in screenshot_frames:
			f = ref_transforms["frames"][int(idx)]
			outname = os.path.join(args.screenshot_dir, os.path.basename(f["file_path"]))
			# skip already generated scene
			if os.path.exists(outname):
				print(f"{outname} exists, skip")
				continue

			cam_matrix = f["transform_matrix"]
			testbed.set_nerf_camera_matrix(np.matrix(cam_matrix)[:-1,:])

			# Some NeRF datasets lack the .png suffix in the dataset metadata
			if not os.path.splitext(outname)[1]:
				outname = outname + ".png"

			print(f"rendering {outname}")
			image = testbed.render(args.width or int(ref_transforms["w"]), args.height or int(ref_transforms["h"]), args.screenshot_spp, True)
			os.makedirs(os.path.dirname(outname), exist_ok=True)
			write_image(outname, image)

	if args.serve:
		server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
		render_pb2_grpc.add_RenderServicer_to_server(
			Renderer(testbed, args.width, args.height, args.screenshot_spp), server)
		server.add_insecure_port('[::]:50051')
		server.start()
		server.wait_for_termination()